import collections
import os
import re
import sqlite3

from sql_metadata import Parser
from transformers import AutoTokenizer


class SqliteTable(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


def get_tables(path_db):
    if not os.path.exists(path_db):
        raise RuntimeError(f"{path_db} not exists")

    # init sqlite connection
    connection = sqlite3.connect(path_db)
    cur = connection.cursor()

    # extract table information
    table_info = parse_db(path_db, cur)
    # TODO: ! add here
    table_names = get_table_names(cur=cur)

    res = list()
    for table_name in table_names:
        # feature
        schema = [_[1] for _ in cur.execute(f'PRAGMA table_info("{table_name}")')]

        # data
        data = None
        # data = cur.execute(f"SELECT * FROM {table_name} LIMIT 5").fetchall()

        # append table
        res.append(
            SqliteTable(
                name=table_name,
                schema=schema,
                data=data,
                table_info=table_info.get(table_name, dict()),
            )
        )

    cur.close()
    return res


def parse_db(path_db, cur=None):
    """Parse the sql file and extract primary and foreign keys

    :param path_file:
    :return:
    """
    table_info = dict()
    table_names = get_table_names(path_db, cur)

    for table_name in table_names:
        pks = get_primary_key(table_name, path_db, cur)
        fks = get_foreign_key(table_name, path_db, cur)

        table_info[table_name] = {"primary_key": pks, "foreign_key": fks}
    return table_info


def execute_query(queries, path_db=None, cur=None):
    """Execute queries and return results. Reuse cur if it's not None."""
    assert not (
        path_db is None and cur is None
    ), "path_db and cur cannot be NoneType at the same time"

    close_in_func = False
    if cur is None:
        con = sqlite3.connect(path_db)
        cur = con.cursor()
        close_in_func = True

    if isinstance(queries, str):
        results = cur.execute(queries).fetchall()
    elif isinstance(queries, list):
        results = list()
        for query in queries:
            res = cur.execute(query).fetchall()
            results.append(res)
    else:
        raise TypeError(f"queries cannot be {type(queries)}")

    # close the connection if needed
    if close_in_func:
        con.close()

    return results


def format_foreign_key(table_name: str, res: list):
    # FROM: self key | TO: target key
    res_clean = list()
    for row in res:
        table, source, to = row[2:5]
        row_clean = f"({table_name}.{source}, {table}.{to})"
        res_clean.append(row_clean)
    return res_clean


def get_foreign_key(table_name, path_db=None, cur=None):
    res_raw = execute_query(f'PRAGMA foreign_key_list("{table_name}")', path_db, cur)
    res = format_foreign_key(table_name, res_raw)
    return res


def get_primary_key(table_name, path_db=None, cur=None):
    res_raw = execute_query(f'PRAGMA table_info("{table_name}")', path_db, cur)
    pks = list()
    for row in res_raw:
        if row[5] == 1:
            pks.append(row[1])
    return pks


def get_table_names(path_db=None, cur=None):
    """Get names of all tables within the database, and reuse cur if it's not None"""
    table_names = execute_query(
        queries="SELECT name FROM sqlite_master WHERE type='table'", path_db=path_db, cur=cur
    )
    table_names = [_[0] for _ in table_names]
    return table_names


def filter_json(raw_response: str) -> str:
    try:
        id_s = raw_response.index("{")
        id_e = raw_response.rindex("}")
        if id_s > id_e:
            raise ValueError("Wrong json format")
        else:
            return raw_response[id_s : id_e + 1]
    except ValueError:
        raise ValueError("Wrong json format")


def get_sql_for_database(path_db=None, cur=None):
    close_in_func = False
    if cur is None:
        con = sqlite3.connect(path_db)
        cur = con.cursor()
        close_in_func = True

    table_names = get_table_names(path_db, cur)

    queries = [f"SELECT sql FROM sqlite_master WHERE tbl_name='{name}'" for name in table_names]

    sqls = execute_query(queries, path_db, cur)

    if close_in_func:
        cur.close()

    return [_[0][0] for _ in sqls]


def get_tokenizer(tokenizer_type: str):
    return 0
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_type, use_fast=False)
    return tokenizer


def count_tokens(string: str, tokenizer_type: str = None, tokenizer=None):
    return 0
    # if tokenizer is None:
    #     tokenizer = get_tokenizer(tokenizer_type)
    #
    # n_tokens = len(tokenizer.encode(string))
    # return n_tokens


def sql_normalization(sql):
    sql = sql.strip()

    def white_space_fix(s):
        parsed_s = Parser(s)
        s = " ".join([token.value for token in parsed_s.tokens])

        return s

    # convert everything except text between single quotation marks to lower case
    def lower(s):
        in_quotation = False
        out_s = ""
        for char in s:
            if in_quotation:
                out_s += char
            else:
                out_s += char.lower()

            if char == "'":
                if in_quotation:
                    in_quotation = False
                else:
                    in_quotation = True

        return out_s

    # remove ";"
    def remove_semicolon(s):
        if s.endswith(";"):
            s = s[:-1]
        return s

    # double quotation -> single quotation
    def double2single(s):
        return s.replace('"', "'")

    def add_asc(s):
        pattern = re.compile(
            r"order by (?:\w+ \( \S+ \)|\w+\.\w+|\w+)(?: (?:\+|\-|\<|\<\=|\>|\>\=) (?:\w+ \( \S+ \)|\w+\.\w+|\w+))*"
        )
        if "order by" in s and "asc" not in s and "desc" not in s:
            for p_str in pattern.findall(s):
                s = s.replace(p_str, p_str + " asc")

        return s

    def sql_split(s):
        while "  " in s:
            s = s.replace("  ", " ")
        s = s.strip()
        i = 0
        toks = []
        while i < len(s):
            tok = ""
            if s[i] == "'":
                tok = tok + s[i]
                i += 1
                while i < len(s) and s[i] != "'":
                    tok = tok + s[i]
                    i += 1
                if i < len(s):
                    tok = tok + s[i]
                    i += 1
            else:
                while i < len(s) and s[i] != " ":
                    tok = tok + s[i]
                    i += 1
                while i < len(s) and s[i] == " ":
                    i += 1
            toks.append(tok)
        return toks

    def remove_table_alias(s):
        tables_aliases = Parser(s).tables_aliases
        new_tables_aliases = {}
        for i in range(1, 11):
            if "t{}".format(i) in tables_aliases.keys():
                new_tables_aliases["t{}".format(i)] = tables_aliases["t{}".format(i)]
        table_names = []
        for tok in sql_split(s):
            if "." in tok:
                table_names.append(tok.split(".")[0])
        for table_name in table_names:
            if table_name in tables_aliases.keys():
                new_tables_aliases[table_name] = tables_aliases[table_name]
        tables_aliases = new_tables_aliases

        new_s = []
        pre_tok = ""
        for tok in sql_split(s):
            if tok in tables_aliases.keys():
                if pre_tok == "as":
                    new_s = new_s[:-1]
                elif pre_tok != tables_aliases[tok]:
                    new_s.append(tables_aliases[tok])
            elif "." in tok:
                split_toks = tok.split(".")
                for i in range(len(split_toks)):
                    if (
                        len(split_toks[i]) > 2
                        and split_toks[i][0] == "'"
                        and split_toks[i][-1] == "'"
                    ):
                        split_toks[i] = split_toks[i].replace("'", "")
                        split_toks[i] = split_toks[i].lower()
                    if split_toks[i] in tables_aliases.keys():
                        split_toks[i] = tables_aliases[split_toks[i]]
                new_s.append(".".join(split_toks))
            else:
                new_s.append(tok)
            pre_tok = tok

        # remove as
        s = new_s
        new_s = []
        for i in range(len(s)):
            if s[i] == "as":
                continue
            if i > 0 and s[i - 1] == "as":
                continue
            new_s.append(s[i])
        new_s = " ".join(new_s)

        # for k, v in tables_aliases.items():
        #     s = s.replace("as " + k + " ", "")
        #     s = s.replace(k, v)

        return new_s

    processing_func = lambda x: remove_table_alias(
        add_asc(lower(white_space_fix(double2single(remove_semicolon(x)))))
    )

    return processing_func(sql.strip())


def sql2skeleton(sql: str, db_schema):
    sql = sql_normalization(sql)

    table_names_original, table_dot_column_names_original, column_names_original = [], [], []
    column_names_original.append("*")
    for table_id, table_name_original in enumerate(db_schema["table_names_original"]):
        table_names_original.append(table_name_original.lower())
        table_dot_column_names_original.append(table_name_original + ".*")
        for column_id_and_name in db_schema["column_names_original"]:
            column_id = column_id_and_name[0]
            column_name_original = column_id_and_name[1]
            table_dot_column_names_original.append(
                table_name_original.lower() + "." + column_name_original.lower()
            )
            column_names_original.append(column_name_original.lower())

    parsed_sql = Parser(sql)
    new_sql_tokens = []
    for token in parsed_sql.tokens:
        # mask table names
        if token.value in table_names_original:
            new_sql_tokens.append("_")
        # mask column names
        elif token.value in column_names_original or token.value in table_dot_column_names_original:
            new_sql_tokens.append("_")
        # mask string values
        elif token.value.startswith("'") and token.value.endswith("'"):
            new_sql_tokens.append("_")
        # mask positive int number
        elif token.value.isdigit():
            new_sql_tokens.append("_")
        # mask negative int number
        elif isNegativeInt(token.value):
            new_sql_tokens.append("_")
        # mask float number
        elif isFloat(token.value):
            new_sql_tokens.append("_")
        else:
            new_sql_tokens.append(token.value.strip())

    sql_skeleton = " ".join(new_sql_tokens)

    # remove JOIN ON keywords
    sql_skeleton = sql_skeleton.replace("on _ = _ and _ = _", "on _ = _")
    sql_skeleton = sql_skeleton.replace("on _ = _ or _ = _", "on _ = _")
    sql_skeleton = sql_skeleton.replace(" on _ = _", "")
    pattern3 = re.compile("_ (?:join _ ?)+")
    sql_skeleton = re.sub(pattern3, "_ ", sql_skeleton)

    # "_ , _ , ..., _" -> "_"
    while "_ , _" in sql_skeleton:
        sql_skeleton = sql_skeleton.replace("_ , _", "_")

    # remove clauses in WHERE keywords
    ops = ["=", "!=", ">", ">=", "<", "<="]
    for op in ops:
        if "_ {} _".format(op) in sql_skeleton:
            sql_skeleton = sql_skeleton.replace("_ {} _".format(op), "_")
    while "where _ and _" in sql_skeleton or "where _ or _" in sql_skeleton:
        if "where _ and _" in sql_skeleton:
            sql_skeleton = sql_skeleton.replace("where _ and _", "where _")
        if "where _ or _" in sql_skeleton:
            sql_skeleton = sql_skeleton.replace("where _ or _", "where _")

    # remove additional spaces in the skeleton
    while "  " in sql_skeleton:
        sql_skeleton = sql_skeleton.replace("  ", " ")

    # double check for order by
    split_skeleton = sql_skeleton.split(" ")
    for i in range(2, len(split_skeleton)):
        if (
            split_skeleton[i - 2] == "order"
            and split_skeleton[i - 1] == "by"
            and split_skeleton[i] != "_"
        ):
            split_skeleton[i] = "_"
    sql_skeleton = " ".join(split_skeleton)

    return sql_skeleton


def isNegativeInt(string):
    if string.startswith("-") and string[1:].isdigit():
        return True
    else:
        return False


def isFloat(string):
    if string.startswith("-"):
        string = string[1:]

    s = string.split(".")
    if len(s) > 2:
        return False
    else:
        for s_i in s:
            if not s_i.isdigit():
                return False
        return True


def jaccard_similarity(skeleton1, skeleton2):
    tokens1 = skeleton1.strip().split(" ")
    tokens2 = skeleton2.strip().split(" ")
    total = len(tokens1) + len(tokens2)

    def list_to_dict(tokens):
        token_dict = collections.defaultdict(int)
        for t in tokens:
            token_dict[t] += 1
        return token_dict

    token_dict1 = list_to_dict(tokens1)
    token_dict2 = list_to_dict(tokens2)

    intersection = 0
    for t in token_dict1:
        if t in token_dict2:
            intersection += min(token_dict1[t], token_dict2[t])
    union = (len(tokens1) + len(tokens2)) - intersection
    return float(intersection) / union



from typing import Optional

import pyarrow as pa

try:
    import adbc_driver_postgresql.dbapi
except ImportError:
    adbc_driver_postgresql = None
from adbc_driver_manager.dbapi import Connection
from pydantic import Field

from aita.datasource.catalog import Catalog
from aita.datasource.sql import SqlDataSource


class PostgreSqlDataSource(SqlDataSource):
    connection_url: str = Field(..., description="Connection URL")

    def __init__(
        self,
        connection_url: Optional[str] = None,
        **kwargs,
    ):
        if not connection_url:
            user = kwargs.get("user")
            password = kwargs.get("password")
            host = kwargs.get("host")
            port = kwargs.get("port")
            database = kwargs.get("database")
            connection_url = self.build_connection_url(
                user=user,
                password=password,
                host=host,
                port=port,
                database=database,
                **kwargs,
            )
        super().__init__(connection_url=connection_url)

    def build_connection_url(
        self,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        database: Optional[str] = None,
        **kwargs,
    ) -> str:
        connection_url = f"postgresql://{user}:{password}@{host}:{port}/{database}"
        if kwargs:
            connection_url += "?" + "&".join([f"{k}={v}" for k, v in kwargs.items()])
        return connection_url

    def connect(self) -> Connection:
        return adbc_driver_postgresql.dbapi.connect(self.connection_url)

    def get_metadata(self, **kwargs):
        with self.connect() as conn:
            catalogs: pa.Table = conn.adbc_get_objects(
                catalog_filter=kwargs.get("database", conn.adbc_current_catalog),
                db_schema_filter=kwargs.get("feature", conn.adbc_current_db_schema),
                table_name_filter=kwargs.get("table", None),
            ).read_all()
            catalog = catalogs.to_pylist()[0]
            return Catalog(**catalog)



from typing import Optional

import pandas as pd
import reflex as rx

from aita import tool
from aita_lab.states.utils import get_model_classes, get_model_fields, get_model_fields_as_dict


class ToolArg(rx.Model):
    name: str
    required: Optional[bool]
    description: Optional[str]
    value: Optional[str]


class Tool(rx.Model):
    id: Optional[str]
    name: str
    args: dict[str, ToolArg]
    description: Optional[str]


class ToolOutput(rx.Base):
    data: pd.DataFrame
    show: bool = False


class ToolState(rx.State):
    tools: list[Tool]

    @rx.var
    def tool_names(self) -> list[str]:
        return [t.name for t in self.tools]

    def load_tools(self):
        self.tools = []
        for t in get_model_classes(tool):
            name, cls = t
            description = get_model_fields(cls, "description")
            args_schema = get_model_fields(cls, "args_schema")
            args = get_model_fields_as_dict(args_schema)
            self.tools.append(
                Tool(
                    name=name,
                    description=description,
                    args={
                        arg["name"]: ToolArg(
                            name=arg["name"],
                            required=arg["required"],
                            description=arg["description"],
                        )
                        for arg in args.values()
                    },
                )
            )

    def on_load(self):
        self.load_tools()



from langchain_core.prompts import ChatPromptTemplate

BasePromptTemplate = ChatPromptTemplate.from_messages(
    messages=[
        (
            "system",
            """
You are an expert in the field of data science, analysis, and data engineering.
""",
        )
    ]
)

BasicContextPromptTemplate = ChatPromptTemplate.from_messages(
    messages=[
        (
            "system",
            """
You are provided with the following context:
{prompt_context}

""",
        )
    ]
)



from typing import Optional

import logging

import adbc_driver_snowflake.dbapi
import pyarrow as pa
from adbc_driver_manager.dbapi import Connection
from langchain_core.pydantic_v1 import Field

from aita.datasource.catalog import Catalog
from aita.datasource.sql import SqlDataSource


class SnowflakeDataSource(SqlDataSource):
    connection_url: str = Field(..., description="Connection URL")

    def __init__(
        self,
        connection_url: Optional[str] = None,
        **kwargs,
    ):
        if not connection_url:
            connection_url = self.build_connection_url(
                **kwargs,
            )
        super().__init__(connection_url=connection_url)

    def build_connection_url(
        self,
        user: str,
        password: str,
        account: str,
        warehouse: Optional[str] = "COMPUTE_WH",
        role: Optional[str] = None,
        database: Optional[str] = None,
        schema: Optional[str] = None,
        **kwargs,
    ) -> str:
        connection_url = f"{user}:{password}@{account}/{database}/{schema}?warehouse={warehouse}"
        if role:
            connection_url += f"&role={role}"
        if kwargs:
            connection_url += "&" + "&".join([f"{k}={v}" for k, v in kwargs.items()])
        return connection_url

    def connect(self) -> Connection:
        logging.info(f"Connecting to Snowflake")
        return adbc_driver_snowflake.dbapi.connect(self.connection_url)

    def get_metadata(self, **kwargs):
        logging.info(f"Getting metadata from Snowflake")
        with self.connect() as conn:
            catalogs: pa.Table = conn.adbc_get_objects(
                catalog_filter=kwargs.get("database", conn.adbc_current_catalog),
                db_schema_filter=kwargs.get("feature", conn.adbc_current_db_schema),
                table_name_filter=kwargs.get("table", None),
            ).read_all()
            catalog = catalogs.to_pylist()[0]
            return Catalog(**catalog)



from typing import Any, Optional

import pyarrow as pa
from langchain_core.pydantic_v1 import Field

from aita.datasource.base import DataSource
from aita.datasource.catalog import Table


class FileDataSource(DataSource):
    type: str = "file"
    file_path: str = Field(..., description="Path to the file")
    file_format: str = Field(..., description="Format of the file")
    file_name: str = Field(..., description="Name of the file")

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, file_path: str, file_format: str, file_name: Optional[str] = None, **kwargs):
        if not file_name:
            file_name = file_path
        super().__init__(
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
        )

    def get_metadata(self, **kwargs) -> Table:
        table_schema = self.to_arrow(**kwargs).schema
        return Table(
            table_name=self.file_name,
            table_columns=[
                {"column_name": name, "column_type": str(table_schema.field(name))}
                for name in table_schema.names
            ],
        )

    def to_arrow(self, **kwargs) -> pa.Table:
        if self.file_format == "csv":
            from pyarrow.csv import read_csv

            return read_csv(self.file_path, **kwargs)
        elif self.file_format == "parquet":
            from pyarrow.parquet import read_table

            return read_table(self.file_path, **kwargs)
        elif self.file_format == "json":
            from pyarrow.json import read_json

            return read_json(self.file_path, **kwargs)
        else:
            raise NotImplementedError(f"FileFormat is unsupported: {self.file_format}")

    def to_pandas(self, **kwargs):
        return self.to_arrow().to_pandas()



from typing import Optional, Union

import base64
import logging
import pickle

import reflex as rx
from langchain_core.messages import AIMessage, ToolMessage
from pandas import DataFrame
from sqlmodel import delete, select

from aita.agent.base import AitaAgent
from aita.agent.factory import AgentFactory
from aita.agent.graph import GraphAgent, ToolMode
from aita_lab.states.base import BaseState
from aita_lab.states.datasource import DataSourceState
from aita_lab.states.prompt_template import PromptTemplate, PromptTemplateState
from aita_lab.states.tool import Tool, ToolArg, ToolOutput
from aita_lab.states.vector_store import VectorStoreState


class QA(rx.Base):
    """A question and answer pair."""

    question: Optional[str]
    answer: Optional[str]


class Chat(rx.Model, table=True):
    """Chat Model"""

    title: str
    user: str
    question: Optional[str]
    answer: Optional[str]
    description: Optional[str]
    created_at: Optional[str]
    updated_at: Optional[str]


DEFAULT_CHATS = {
    "Intros": [],
}


class ChatState(BaseState):
    """The app state."""

    # chat states
    chats: dict[str, list[QA]] = DEFAULT_CHATS

    current_chat = "Intros"

    question: str

    new_chat_title: str = ""

    # basic model states
    current_chat_model: str

    current_datasource: str

    # RAG states
    current_embedding_model: Optional[str] = ""

    current_prompt_template: Optional[PromptTemplate] = None

    current_k_shot: int = 0

    current_vector_feature: str = ""

    # agent states
    current_chat_agent_type: str = ""

    _current_chat_agent: Optional[Union[AitaAgent, GraphAgent]] = None

    _current_embedding_agent: Optional[AitaAgent] = None

    _current_chat_agent_state_change: bool = False

    _current_embedding_agent_state_change: bool = False

    # tool states
    current_tools: dict[str, Tool] = {}

    current_tool_id: Optional[str] = None

    run_tool_output: Optional[ToolOutput] = None

    # Whether we are processing the question.
    processing: bool = False

    vector_feature_dialog_open: bool = False

    @rx.var
    def current_tools_as_list(self) -> list[Tool]:
        if self.current_tools:
            return list(self.current_tools.values())
        return []

    @rx.var
    def current_tool_name(self) -> str:
        if self.current_tools and self.current_tool_id:
            return self.current_tools[self.current_tool_id].name
        return ""

    @rx.var
    def current_tool_args(self) -> list[ToolArg]:
        if self.current_tools and self.current_tool_id:
            return list(self.current_tools[self.current_tool_id].args.values())
        return []

    def close_vector_feature_dialog(self):
        self.vector_feature_dialog_open = False

    def set_current_chat_model(self, chat_model: str):
        if self.current_chat_model != chat_model:
            self.current_chat_model = chat_model
            self._current_chat_agent_state_change = True

    def set_current_datasource(self, datasource: str):
        if self.current_datasource != datasource:
            self.current_datasource = datasource
            self._current_chat_agent_state_change = True

    def set_current_chat_agent_type(self, chat_agent_type: str):
        if self.current_chat_agent_type != chat_agent_type:
            self.current_chat_agent_type = chat_agent_type
            self._current_chat_agent_state_change = True

    def set_current_prompt_template(self, prompt_template_name: str):
        self.current_prompt_template = PromptTemplateState.get_prompt_template(prompt_template_name)
        if self.current_prompt_template and self.current_prompt_template.k_shot > 0:
            self.vector_feature_dialog_open = True

    def _create_chat_agent(self, **kwargs):
        if not self._current_chat_agent or self._current_chat_agent_state_change:
            logging.info(
                f"Creating {self.current_chat_agent_type} agent with model {self.current_chat_model}"
            )

            self._current_chat_agent = AgentFactory.create_agent(
                agent_type=self.current_chat_agent_type,
                model=self.current_chat_model,
                **kwargs,
            )
            if self.current_datasource:
                datasource = DataSourceState.connect(self.current_datasource)
                self._current_chat_agent.add_datasource(datasource)

            logging.info(
                f"Created {self.current_chat_agent_type} agent with model {self.current_chat_model}"
            )
            self._current_chat_agent_state_change = False

    def set_current_embedding_model(self, embedding_model: str):
        if self.current_embedding_model != embedding_model:
            self.current_embedding_model = embedding_model
            self._current_embedding_agent_state_change = True

    def _create_embedding_agent(self):
        if not self._current_embedding_agent or self._current_embedding_agent_state_change:
            logging.info(f"Creating embedding agent with model {self.current_embedding_model}")
            self._current_embedding_agent = AgentFactory.create_agent(
                agent_type="embedding", model=self.current_embedding_model
            )
            self._current_embedding_agent_state_change = False

    def should_create_embedding_agent(self):
        if self.current_prompt_template and self.current_prompt_template.k_shot > 0:
            return True

    def set_current_tool_by_id(self, tool_id: str):
        if tool_id:
            logging.info(f"Setting current tool by id to {tool_id}")
            self.current_tool_id = tool_id

    def set_current_tool_by_name(self, tool_name: str):
        if tool_name:
            logging.info(f"Setting current tool by name to {tool_name}")
            for tool_id, tool in self.current_tools.items():
                if tool.name == tool_name:
                    self.current_tool_id = tool_id
                    break

    def update_current_tool_arg(self, key: str, value: str):
        tool_arg = self.current_tools[self.current_tool_id].args[key]
        new_tool_arg = ToolArg(
            name=key, reqired=tool_arg.required, description=tool_arg.description, value=value
        )
        self.current_tools[self.current_tool_id].args[key] = new_tool_arg

    def delete_current_tool(self):
        if self.current_tool_id:
            logging.info(f"Deleting current tool {self.current_tool_id}")
            del self.current_tools[self.current_tool_id]
            self.current_tool_id = None

    def cancel_tool(self):
        logging.info(f"Canceling tool {self.current_tool_id}")
        self._current_chat_agent.cancel_tool(self.current_tool_id)

    def create_chat(self):
        """Create a new chat."""
        # Add the new chat to the list of chats.
        self.current_chat = self.new_chat_title
        self.chats[self.new_chat_title] = []
        with rx.session() as session:
            session.add(
                Chat(
                    title=self.new_chat_title,
                    user=self.user.name,
                    question=None,
                    answer=None,
                    description=None,
                    created_at=None,
                    updated_at=None,
                )
            )
            session.commit()

    def delete_chat(self):
        """Delete the current chat."""
        with rx.session() as session:
            session.exec(delete(Chat).where(Chat.title == self.current_chat))
            session.commit()
        del self.chats[self.current_chat]
        self.load_chats()
        self.current_chat = list(self.chats.keys())[0]

    def set_chat(self, chat_title: str):
        """Set the title of the current chat.

        Args:
            chat_title: The name of the chat.
        """
        self.current_chat = chat_title

    @rx.var
    def chat_titles(self) -> list[str]:
        """Get the list of chat titles.

        Returns:
            The list of chat names.
        """
        return list(self.chats.keys())

    def _commit_chat(self, title: str, question: str, answer: str):
        with rx.session() as session:
            session.add(
                Chat(
                    title=title,
                    user=self.user.name,
                    question=question,
                    answer=answer,
                    description=None,
                    created_at=None,
                    updated_at=None,
                )
            )
            session.commit()

    def _process_agent_response(self, events: list[dict]):
        for event in events:
            if hasattr(event, "content"):
                message = event
            else:
                message = event["messages"][-1]
            if isinstance(message, AIMessage):
                self.chats[self.current_chat][-1].answer += message.content
            if isinstance(message, ToolMessage):
                if message.artifact is not None:
                    encoded_artifact = message.artifact.encode("utf-8")
                    artifact = base64.b64decode(encoded_artifact)
                    result = pickle.loads(artifact)
                    if isinstance(result, DataFrame):
                        self.run_tool_output = ToolOutput(data=result, show=True)
            yield

    def _process_agent_state(self):
        chat_state = self._current_chat_agent.get_current_state()
        if chat_state and chat_state.next:
            # having an action to execute
            for tool_call in self._current_chat_agent.get_current_tool_calls():
                tool = Tool(
                    id=tool_call["id"],
                    name=tool_call["name"],
                    args={
                        key: ToolArg(name=key, value=str(value))  # TODO handle other types
                        for key, value in tool_call["args"].items()
                    },
                )
                self.current_tools[tool_call["id"]] = tool
                if not self.current_tool_id:
                    self.current_tool_id = tool_call["id"]

            # Add the tool call to the answer
            self.chats[self.current_chat][
                -1
            ].answer += f"In order to assist you further, I need to run a tool shown in the kernel. Please confirm to run the tool."

    async def run_tool(self):
        self.processing = True

        tool_args = {
            tool_arg.name: tool_arg.value
            for key, tool_arg in self.current_tools[self.current_tool_id].args.items()
        }
        logging.info("Updating tool args: " + str(tool_args.items()))
        self._current_chat_agent.update_tool(self.current_tool_id, tool_args)

        # Add a empty question to the list of questions.
        qa = QA(question="", answer="")
        self.chats[self.current_chat].append(qa)

        events = self._current_chat_agent.stream(tool_mode=ToolMode.ONCE, display=False)
        for event in self._process_agent_response(events):
            yield

        self._process_agent_state()

        # Toggle the processing flag.
        self.processing = False

        # commit the chat to the database
        self._commit_chat(
            self.current_chat,
            self.chats[self.current_chat][-1].question,
            self.chats[self.current_chat][-1].answer,
        )

    async def process_question(self, form_data: dict[str, str]):

        # Clear the input and start the processing.
        self.processing = True

        # Get the question from the form
        question = form_data["question"]

        # Check if the question is empty
        if question == "":
            return

        # Add the question to the list of questions.
        qa = QA(question=question, answer="")
        self.chats[self.current_chat].append(qa)

        yield

        logging.info(f"Processing question: {question}")

        # init the chat agent
        self._create_chat_agent()

        # init the embedding agent
        if self.should_create_embedding_agent():
            self._create_embedding_agent()

            # embed the question
            embedded_question = self._current_embedding_agent.embed(question)

            # retrieve similar features
            top_k_features = VectorStoreState._retrieve_vector_features(
                self.current_vector_feature, embedded_question, self.current_k_shot
            )

            # TODO: add similar features to the prompt template
            self._current_chat_agent.add_prompt_context(top_k_features)

        async for value in self._invoke_agent(question):
            yield

        # Toggle the processing flag.
        self.processing = False

        # commit the chat to the database
        self._commit_chat(
            self.current_chat,
            self.chats[self.current_chat][-1].question,
            self.chats[self.current_chat][-1].answer,
        )

    async def _invoke_agent(self, question: Optional[str] = ""):
        """Get the response from the API.

        Args:
            question: The question to ask the API.
        """

        # Get the response and add it to the answer.
        events = self._current_chat_agent.stream(question, display=False)
        for event in self._process_agent_response(events):
            yield

        self._process_agent_state()

    def load_chats(self):
        """Load the chats from the database."""
        with rx.session() as session:
            chats = session.exec(select(Chat).where(Chat.user == self.user.name)).all()
            for chat in chats:
                self.chats[chat.title] = [QA(question=chat.question, answer=chat.answer)]
            if not self.chats:
                self.chats = DEFAULT_CHATS

    def on_load(self):
        self.load_chats()



from __future__ import annotations

import reflex as rx


def table(tabular_data: list[list]):
    return rx.table.root(
        rx.table.header(
            rx.table.row(
                *[rx.table.column_header_cell(cell) for cell in tabular_data[0]],
            ),
        ),
        rx.table.body(
            *[
                rx.table.row(
                    *[
                        rx.table.row_header_cell(cell) if i == 0 else rx.table.cell(cell)
                        for i, cell in enumerate(row)
                    ],
                )
                for row in tabular_data[1:]
            ],
        ),
    )



from aita.prompt.enums import ExampleType, ReprType, SelectorType
from aita.prompt.ExampleFormatTemplate import *
from aita.prompt.ExampleSelectorTemplate import *
from aita.prompt.PromptICLTemplate import BasicICLPrompt
from aita.prompt.PromptReprTemplate import *


def get_repr_cls(repr_type: str):
    if repr_type == ReprType.CODE_REPRESENTATION:
        repr_cls = SQLPrompt
    elif repr_type == ReprType.TEXT_REPRESENTATION:
        repr_cls = TextPrompt
    elif repr_type == ReprType.OPENAI_DEMOSTRATION:
        repr_cls = NumberSignPrompt
    elif repr_type == ReprType.BASIC:
        repr_cls = BaselinePrompt
    elif repr_type == ReprType.ALPACA_SFT:
        repr_cls = InstructionPrompt
    elif repr_type == ReprType.OPENAI_DEMOSTRATION_WFK:
        repr_cls = NumberSignWithForeignKeyPrompt
    elif repr_type == ReprType.BASIC_WOFK:
        repr_cls = BaselineWithoutForeignKeyPrompt
    elif repr_type == ReprType.TEXT_REPRESENTATION_WFK:
        repr_cls = TextWithForeignKeyPrompt
    elif repr_type == ReprType.ALPACA_SFT_WFK:
        repr_cls = InstructionWithForeignKeyPrompt
    elif repr_type == ReprType.OPENAI_DEMOSTRATION_WORULE:
        repr_cls = NumberSignWithoutRulePrompt
    elif repr_type == ReprType.CODE_REPRESENTATION_WRULE:
        repr_cls = SQLWithRulePrompt
    elif repr_type == ReprType.ALPACA_SFT_WRULE:
        repr_cls = InstructionWithRulePrompt
    elif repr_type == ReprType.TEXT_REPRESENTATION_WRULE:
        repr_cls = TextWithRulePrompt
    elif repr_type == ReprType.CODE_REPRESENTATION_COT:
        repr_cls = SQLCOTPrompt
    elif repr_type == ReprType.TEXT_REPRESENTATION_COT:
        repr_cls = TextCOTPrompt
    elif repr_type == ReprType.OPENAI_DEMOSTRATION_COT:
        repr_cls = NumberSignCOTPrompt
    elif repr_type == ReprType.ALPACA_SFT_COT:
        repr_cls = InstructionCOTPrompt
    elif repr_type == ReprType.CBR:
        repr_cls = CBRPrompt
    else:
        raise ValueError(f"{repr_type} is not supported yet")
    return repr_cls


def get_example_format_cls(example_format: str):
    if example_format == ExampleType.ONLY_SQL:
        example_format_cls = SqlExampleStyle
    elif example_format == ExampleType.QA:
        example_format_cls = QuestionSqlExampleStyle
    elif example_format == ExampleType.QAWRULE:
        example_format_cls = QuestionSqlWithRuleExampleStyle
    elif example_format == ExampleType.COMPLETE:
        example_format_cls = CompleteExampleStyle
    elif example_format == ExampleType.OPENAI_DEMOSTRATION_QA:
        example_format_cls = NumberSignQuestionSqlExampleStyle
    elif example_format == ExampleType.BASIC_QA:
        example_format_cls = BaselineQuestionSqlExampleStyle
    else:
        raise ValueError(f"{example_format} is not supported yet!")
    return example_format_cls


def get_example_selector(selector_type: str):
    if selector_type == SelectorType.COS_SIMILAR:
        selector_cls = CosineSimilarExampleSelector
    elif selector_type == SelectorType.RANDOM:
        selector_cls = RandomExampleSelector
    elif selector_type == SelectorType.EUC_DISTANCE:
        selector_cls = EuclideanDistanceExampleSelector
    elif selector_type == SelectorType.EUC_DISTANCE_THRESHOLD:
        selector_cls = EuclideanDistanceThresholdExampleSelector
    elif selector_type == SelectorType.EUC_DISTANCE_PRE_SKELETON_SIMILARITY_THRESHOLD:
        selector_cls = EuclideanDistancePreSkeletonSimilarThresholdSelector
    elif selector_type == SelectorType.EUC_DISTANCE_PRE_SKELETON_SIMILARITY_PLUS:
        selector_cls = EuclideanDistancePreSkeletonSimilarPlusSelector
    else:
        raise ValueError(f"{selector_type} is not supported yet!")
    return selector_cls


def prompt_factory(repr_type: str, k_shot: int, example_format: str, selector_type: str):
    repr_cls = get_repr_cls(repr_type)

    if k_shot == 0:
        assert repr_cls is not None
        cls_name = f"{repr_type}_{k_shot}-SHOT"

        class PromptClass(repr_cls, BasicICLPrompt):
            name = cls_name
            NUM_EXAMPLE = k_shot

            def __init__(self, *args, **kwargs):
                repr_cls.__init__(self, *args, **kwargs)
                # init tokenizer
                BasicICLPrompt.__init__(self, *args, **kwargs)

    else:
        example_format_cls = get_example_format_cls(example_format)
        selector_cls = get_example_selector(selector_type)
        cls_name = f"{repr_type}_{k_shot}-SHOT_{selector_type}_{example_format}-EXAMPLE"

        class PromptClass(selector_cls, example_format_cls, repr_cls, BasicICLPrompt):
            name = cls_name
            NUM_EXAMPLE = k_shot

            def __init__(self, *args, **kwargs):
                selector_cls.__init__(self, *args, **kwargs)
                # init tokenizer
                BasicICLPrompt.__init__(self, *args, **kwargs)

    return PromptClass



"""The home page of the app."""

import reflex as rx

from aita.datasource.factory import get_supported_datasources
from aita_lab.states.datasource import DataSource, DataSourceState
from aita_lab.templates import template


def show_datasource(datasource: DataSource):
    return rx.table.row(
        rx.table.cell(datasource.name),
        rx.table.cell(datasource.connection_url),
        rx.table.cell(update_datasource(datasource)),
        rx.table.cell(
            rx.button(
                "Delete",
                on_click=lambda: DataSourceState.delete_datasource(datasource),
            )
        ),
    )


def add_datasource():
    datasources = [ds.name for ds in get_supported_datasources()]
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.button(
                rx.flex("Add Data Source", rx.icon(tag="plus", width=24, height=24), spacing="3"),
                size="4",
                radius="full",
                on_click=DataSourceState.toggle_dialog,
            ),
        ),
        rx.dialog.content(
            rx.dialog.title(
                "Connection Information",
                size="1",
                font_family="Inter",
                padding_top="1em",
            ),
            rx.dialog.description(
                "Select the data source and enter the connection information for your data source",
                size="2",
                mb="4",
                padding_bottom="1em",
            ),
            rx.flex(
                rx.text(
                    "Name *",
                    as_="div",
                    size="2",
                    mb="1",
                    weight="bold",
                ),
                rx.input(
                    placeholder="Enter the name of the data source",
                    on_blur=DataSourceState.set_name,
                    required=True,
                ),
                rx.text(
                    "Type *",
                    as_="div",
                    size="2",
                    mb="1",
                    weight="bold",
                ),
                rx.select(
                    datasources,
                    placeholder="Select the data source type",
                    on_change=DataSourceState.set_datasource_type,
                ),
                rx.cond(
                    DataSourceState.datasource_type,
                    rx.vstack(
                        rx.foreach(
                            DataSourceState.datasource_attributes,
                            lambda field: rx.vstack(
                                rx.text(
                                    field,
                                    as_="div",
                                    size="2",
                                    mb="1",
                                    weight="bold",
                                ),
                                rx.input(
                                    placeholder=f"Enter the {field} for the data source",
                                    on_blur=lambda x: DataSourceState.set_datasource_attributes(
                                        field, x
                                    ),
                                    width="100%",
                                ),
                                width="100%",
                            ),
                        ),
                        width="100%",
                    ),
                ),
                direction="column",
                spacing="4",
            ),
            rx.cond(
                DataSourceState.missing_attributes,
                rx.chakra.alert(
                    rx.chakra.alert_icon(),
                    rx.chakra.alert_title(
                        "Please fill in all the required fields",
                    ),
                    status="warning",
                    mb="3",
                    mt="3",
                ),
            ),
            rx.flex(
                rx.switch(
                    checked=DataSourceState.allow_crawl_catalog,
                    on_change=DataSourceState.change_crawl_catalog,
                ),
                rx.text("Crawl Catalog", padding_top="4px"),
                spacing="2",
            ),
            rx.flex(
                rx.dialog.close(
                    rx.button(
                        "Cancel",
                        variant="soft",
                        color_scheme="gray",
                        on_click=DataSourceState.toggle_dialog,
                    )
                ),
                rx.dialog.close(
                    rx.button(
                        "Connect",
                        on_click=DataSourceState.connect_and_add_datasource,
                        variant="solid",
                    )
                ),
                padding_top="1em",
                spacing="3",
                mt="4",
                justify="end",
            ),
            style={"width": 450},
            box_shadow="lg",
            padding="1em",
            border_radius="25px",
            font_family="Inter",
        ),
        open=DataSourceState.is_open,
    )


def update_datasource(datasource: DataSource):
    return rx.dialog.root(
        rx.dialog.trigger(
            rx.button(
                rx.icon("square_pen", width=24, height=24),
                on_click=lambda: DataSourceState.set_datasource(datasource),
            ),
        ),
        rx.dialog.content(
            rx.dialog.title("Data Source Details"),
            rx.dialog.description(
                "Update your data source details.",
                size="2",
                mb="4",
                padding_bottom="1em",
            ),
            rx.flex(
                rx.text(
                    "Name",
                    as_="div",
                    size="2",
                    mb="1",
                    weight="bold",
                ),
                rx.input(
                    placeholder=datasource.name,
                    default_value=datasource.name,
                    on_blur=DataSourceState.set_name,
                ),
                rx.text(
                    "Connection URL",
                    as_="div",
                    size="2",
                    mb="1",
                    weight="bold",
                ),
                rx.input(
                    placeholder=datasource.connection_url,
                    default_value=datasource.connection_url,
                    on_blur=DataSourceState.set_connection_url,
                ),
                direction="column",
                spacing="3",
            ),
            rx.flex(
                rx.dialog.close(
                    rx.button(
                        "Cancel",
                        variant="soft",
                        color_scheme="gray",
                    ),
                ),
                rx.dialog.close(
                    rx.button(
                        "Update",
                        on_click=DataSourceState.update_datasource,
                        variant="solid",
                    ),
                ),
                padding_top="1em",
                spacing="3",
                mt="4",
                justify="end",
            ),
            style={"max_width": 450},
            box_shadow="lg",
            padding="1em",
            border_radius="25px",
        ),
    )


def content_grid():
    return rx.fragment(
        rx.vstack(
            rx.box(
                add_datasource(),
            ),
            rx.divider(),
            rx.hstack(
                rx.heading(
                    f"Total: {DataSourceState.num_datasources} datasources",
                    size="5",
                    font_family="Inter",
                ),
                rx.spacer(),
                rx.select(
                    ["name", "type", "connection_url"],
                    placeholder="Sort By: Name",
                    size="3",
                    on_change=lambda sort_value: DataSourceState.sort_values(sort_value),
                    font_family="Inter",
                ),
                width="100%",
                padding_top="2em",
                padding_bottom="1em",
            ),
            rx.table.root(
                rx.table.header(
                    rx.table.row(
                        rx.table.column_header_cell("Name"),
                        rx.table.column_header_cell("Connection URL"),
                        rx.table.column_header_cell("Edit"),
                        rx.table.column_header_cell("Delete"),
                    ),
                ),
                rx.table.body(rx.foreach(DataSourceState.datasources, show_datasource)),
                # variant="surface",
                size="3",
                width="100%",
                justify="stretch",
            ),
        ),
    )


@template(route="/datasource", title="Data Source", on_load=DataSourceState.on_load())
def datasource() -> rx.Component:
    """The home page.

    Returns:
        The UI for the home page.
    """
    return rx.vstack(
        rx.heading("Data Source", size="8"),
        rx.text("Connect to your data source"),
        rx.box(
            content_grid(),
            margin_top="20px",
            width="100%",
        ),
        # make the page full width
        width="100%",
    )



from typing import Any, Dict, Literal, Optional, Sequence, Type, Union

import base64
import pickle
from abc import ABC

import pandas as pd
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_core.tools import BaseTool
from sqlalchemy.engine import Result

from aita.datasource.sql import SqlDataSource


class SqlDataSourceTool(BaseTool, ABC):
    datasource: Optional[SqlDataSource] = Field(None, exclude=True)

    class Config(BaseTool.Config):
        pass


class QueryInput(BaseModel):
    query: str = Field(description="sql query that can be executed by the sql database.")


class SqlQueryTool(SqlDataSourceTool):
    """Tool for querying a SQL database."""

    name: str = "sql_database_query"
    description: str = """
    Execute a SQL query against the database and get back the result..
    If the query is not correct, an error message will be returned.
    If an error is returned, rewrite the query, check the query, and try again.
    """
    args_schema: Type[BaseModel] = QueryInput
    response_format: Literal["content", "content_and_artifact"] = "content_and_artifact"

    def _run(
        self,
        query,
        **kwargs,
    ) -> (str, str):
        """Execute the query, return the results or an error message."""
        try:
            result = self.datasource.to_pandas(query)

            # Serialize the result to a base64 encoded string as the artifact
            artifact = base64.b64encode(pickle.dumps(result)).decode("utf-8")
            return result, artifact
        except Exception as e:
            return "Received an error while executing the query: {}".format(str(e)), ""


class Column(BaseModel):
    column_name: str = Field(..., description="Name of the column")
    column_type: str = Field(..., description="Type of the column")
    nullable: Optional[bool] = Field(None, description="Whether the column is nullable")
    primary_key: Optional[bool] = Field(None, description="Whether the column is a primary key")


class Table(BaseModel):
    table_name: str = Field(..., description="Name of the table")
    table_columns: Sequence[Column] = Field(..., description="List of columns in the table")
    table_type: Optional[str] = Field(..., description="Type of the table")


class CreateTableTool(SqlDataSourceTool):
    """Tool for creating a table in a SQL database."""

    datasource: Optional[SqlDataSource] = Field(None, exclude=True)
    name: str = "create_table"
    description: str = """
    Create a table in the database.
    If the table already exists, an error message will be returned.
    input arguments are table_name and table_columns.
    """
    args_schema: Type[BaseModel] = Table

    def _run(
        self,
        table_name: str,
        table_columns: Sequence[Column],
        **kwargs,
    ) -> Union[str, Sequence[Dict[str, Any]], Result]:
        """Execute the query, return the results or an error message."""
        columns = ",\n".join(
            f'{column.column_name} "{column.column_type}"' for column in table_columns
        )
        return self.datasource.execute(
            "CREATE TABLE {table_name} ({columns})".format(table_name=table_name, columns=columns)
        )


class ConvertToPandasTool(SqlDataSourceTool):
    """Tool for converting a SQL query result to a pandas dataframe."""

    name: str = "convert_to_pandas"
    description: str = """
    Convert a SQL query result to a pandas dataframe.
    The tool can only be used if the data source is provided.
    If the query result is not correct, an error message will be returned.
    """
    args_schema: Type[BaseModel] = QueryInput

    def _run(
        self,
        query: str,
        **kwargs,
    ) -> pd.DataFrame:
        """Execute the query, return the result as pandas or an error message."""
        return self.datasource.to_pandas(query)


class ConvertToArrowTool(SqlDataSourceTool):
    """Tool for converting a SQL query result to a PyArrow Table."""

    name: str = "convert_to_arrow"
    description: str = """
    Convert a SQL query result to a PyArrow Table.
    If the query result is not correct, an error message will be returned.
    """
    args_schema: Type[BaseModel] = QueryInput

    def _run(
        self,
        query: str,
        **kwargs,
    ) -> Table:
        """Execute the query, return the results or an error message."""
        return self.datasource.to_arrow(query)



from typing import Dict, Optional

from aita.agent.graph import GraphAgent
from aita.tool.sql_tool import (
    ConvertToArrowTool,
    ConvertToPandasTool,
    CreateTableTool,
    SqlQueryTool,
)


class SqlAgent(GraphAgent):
    type: str = "sql"
    description: str = "A SQL agent that can use SQL Tools to interact with SQL databases."

    def __init__(
        self,
        model: str,
        model_parameters: Optional[Dict] = None,
    ):
        super().__init__(
            [SqlQueryTool(), CreateTableTool(), ConvertToArrowTool()],
            model,
            model_parameters,
        )



class SqlExampleStyle(object):
    """Only show sqls as examples"""

    def get_example_prefix(self):
        return "/* Some SQL examples are provided based on similar problems: */\n"

    def format_example(self, example: dict):
        return example["query"]


class QuestionSqlExampleStyle(object):
    """Provide QA pair as examples"""

    def get_example_prefix(self):
        return "/* Some SQL examples are provided based on similar problems: */\n"

    def format_example(self, example: dict):
        template_qa = "/* Answer the following: {} */\n{}"
        return template_qa.format(example["question"], example["query"])


class QuestionSqlWithRuleExampleStyle(object):
    """Provide QA pair as examples"""

    def get_example_prefix(self):
        return "/* Some SQL examples are provided based on similar problems: */\n"

    def format_example(self, example: dict):
        template_qa = "/* Answer the following with no explanation: {} */\n{}"
        return template_qa.format(example["question"], example["query"])


class CompleteExampleStyle(object):
    """Examples are in the same format as target question"""

    def get_example_prefix(self):
        return ""

    def format_example(self, example: dict):
        return f"{self.format_question(example)}\n{example['query']}"


class NumberSignQuestionSqlExampleStyle(object):
    """
    Provide QA pair as examples
    """

    def get_example_prefix(self):
        return "### Some example pairs of question and corresponding SQL query are provided based on similar problems:\n\n"

    def format_example(self, example: dict):
        template_qa = "### {}\n{}"
        return template_qa.format(example["question"], example["query"])


class BaselineQuestionSqlExampleStyle(object):
    """
    Provide QA pair as examples
    """

    def get_example_prefix(self):
        return ""

    def format_example(self, example: dict):
        template_qa = "Example Q: {}\nExample A: {}"
        return template_qa.format(example["question"], example["query"])



from typing import List, Optional

from pydantic import BaseModel, Field


class Column(BaseModel):
    column_name: str = Field(..., description="Name of the column")
    xdbc_type_name: Optional[str] = Field(
        None, description="Type of the column", alias="column_type"
    )
    xdbc_nullable: Optional[bool] = Field(
        None, description="Whether the column is nullable", alias="nullable"
    )
    primary_key: Optional[bool] = Field(None, description="Whether the column is a primary key")

    class Config:
        populate_by_name = True


class Table(BaseModel):
    table_name: str = Field(..., description="Name of the table")
    table_columns: List[Column] = Field(..., description="List of columns in the table")
    table_type: Optional[str] = Field(..., description="Type of the table")

    class Config:
        populate_by_name = True


class Schema(BaseModel):
    db_schema_name: str = Field(
        ..., description="Name of the database feature", alias="schema_name"
    )
    db_schema_tables: List[Table] = Field(
        ..., description="List of tables in the feature", alias="tables"
    )

    class Config:
        populate_by_name = True


class Catalog(BaseModel):
    catalog_name: str = Field(..., description="Name of the catalog", alias="catalog_name")
    catalog_db_schemas: Optional[List[Schema]] = Field(
        None, description="List of database schemas in the catalog", alias="schemas"
    )
    catalog_tables: Optional[List[Table]] = Field(
        None, description="List of tables in the catalog", alias="tables"
    )
    catalog_columns: Optional[List[Column]] = Field(
        None, description="List of columns in the catalog", alias="columns"
    )

    class Config:
        populate_by_name = True



from typing import Type

from langchain_core.pydantic_v1 import BaseModel, Field

from aita.tool.python_tool import PythonTool


class ArrowInput(BaseModel):
    script: str = Field(description="PyArrow analysis script")


class ArrowTool(PythonTool):
    """Tool for using Apache Arrow in Python."""

    name: str = "pyarrow_tool"
    description: str = """
    Apache Arrow is a cross-language development platform for in-memory data analysis.
    This tool allows you to run PyArrow script in Python.

    PyArrow Table is available in the script context.
    """
    args_schema: Type[BaseModel] = ArrowInput



from .agent import agent
from .catalog import catalog
from .chat import chat
from .datasource import datasource
from .prompt_template import prompt_template
from .settings import settings
from .tool import tool
from .vector_store import vector_store



from typing import Any, Optional

import reflex as rx

from aita_lab.states.llm_providers import ChatModelProvider, EmbeddingModelProvider


def select_model(
    model_provider,
    model_value,
    on_model_value_change,
    style: Optional[dict] = None,
) -> rx.Component:
    return rx.select.root(
        rx.select.trigger(
            placeholder="Select a model",
            style=style if style else {"width": "100%"},
        ),
        rx.select.content(
            *[
                rx.select.group(
                    rx.select.label(p.value.name),
                    rx.foreach(
                        p.value.models,
                        lambda x: rx.select.item(x, value=f"{p.value.id}:{x}"),
                    ),
                    width="100%",
                )
                for p in list(model_provider)
            ],
        ),
        value=model_value,
        on_change=on_model_value_change,
        default_value="gpt-3.5-turbo",
    )


def select_chat_model(
    model: str,
    set_model: Any,
    style: Optional[dict] = None,
) -> rx.Component:
    """The model selector."""
    return rx.form(
        rx.chakra.form_control(
            rx.text(
                "Chat Model *",
                asi_="div",
                mb="1",
                size="2",
                weight="bold",
            ),
            select_model(ChatModelProvider, model, set_model, style),
            label="Model",
            width="100%",
        ),
        width="100%",
    )


def select_embedding_model(model, set_model) -> rx.Component:
    """The model selector."""
    return rx.form(
        rx.chakra.form_control(
            rx.text(
                "Embedding Model *",
                asi_="div",
                mb="1",
                size="2",
                weight="bold",
            ),
            select_model(EmbeddingModelProvider, model, set_model),
            label="Model",
            width="100%",
        ),
        width="100%",
    )



from typing import Dict, Literal, Optional, Union

from langchain_core.messages import ToolMessage
from langchain_core.output_parsers.openai_tools import PydanticToolsParser
from langchain_core.prompts import PromptTemplate
from langchain_core.pydantic_v1 import BaseModel, ValidationError
from langchain_core.runnables import RunnableSerializable
from langchain_core.runnables.config import RunnableConfig
from langgraph.graph import END, StateGraph

from aita.agent.graph import GraphAgent
from aita.states import MessageState


class ValidatorAgent(GraphAgent):
    validator: BaseModel
    validator_chain: RunnableSerializable

    def __int__(
        self,
        validator: BaseModel,
        model: Union[RunnableSerializable, str],
        model_parameters: Optional[Dict] = None,
        **kwargs,
    ):
        self.validator_tool = validator

        self.base_prompt_template = PromptTemplate(
            template="Please return the output given messages and following the feature validation\n{messages}",
            input_variables=["messages"],
        )

        super().__init__([self.validator_tool], model, model_parameters, **kwargs)

    def call_model(self, state: list, config: RunnableConfig):
        messages = {**state, "user_info": config.get("user_id", None)}
        response = self.chain.invoke(messages, config)
        return {"messages": [response]}

    def validate(self, state: list, config: RunnableConfig):
        messages = state["messages"][-1]
        try:
            response = self.validator.invoke(messages)
            return {"messages": [response]}
        except ValidationError as e:
            state = state + [
                ToolMessage(
                    content=f"{repr(e)}\n\nPay close attention to the function feature.\n\n"
                    + self.validator.schema_json()
                    + " Respond by fixing all validation errors.",
                    tool_call_id=response.tool_calls[0]["id"],
                ),
            ]
            return state

    def _build_graph(self):

        workflow = StateGraph(MessageState)

        workflow.add_node("agent", self.call_model)
        workflow.add_node("tools", self.build_tool_node(self.tools))
        workflow.add_node("validator", self.call_model)
        workflow.add_node("validation", self._validate)

        workflow.add_conditional_edges("agent", self._should_call_tool)
        workflow.add_conditional_edges("validator", self._should_validate)
        workflow.add_edge("tools", "agent")
        workflow.add_edge("validation", "validator")

        workflow.set_entry_point("agent")
        return workflow.compile(checkpointer=self.memory, interrupt_before=["validator", "tools"])

    def _validate(self, state: list, config: RunnableConfig):
        message = state["messages"][-1]
        try:
            validator = PydanticToolsParser(tools=[self.validator])
            validator.invoke(message)
            response = self.validator_chain.invoke(config)
            response.additional_kwargs["validated"] = True
            return {"messages": [response]}
        except ValidationError as e:
            return {
                "messages": [
                    ToolMessage(
                        content=f"{repr(e)}\n\nPay close attention to the function feature.\n\n"
                        + " Respond by fixing all validation errors.",
                        tool_call_id=message.tool_calls[0]["id"],
                    )
                ]
            }

    def _should_call_tool(self, state: list) -> Literal["tools", "validator"]:
        if isinstance(state, list):
            ai_message = state[-1]
        elif messages := state.get("messages", []):
            ai_message = messages[-1]
        else:
            raise ValueError(f"No messages found in input state to tool_edge: {state}")
        if hasattr(ai_message, "tool_calls") and len(ai_message.tool_calls) > 0:
            return "tools"
        return "validator"

    def _should_validate(self, state: list) -> Literal["agent", "validation", "__end__"]:
        if (
            state["messages"][-1].tool_calls
            and "validated" not in state["messages"][-1].additional_kwargs
        ):
            return "validation"
        return END



import pandas as pd

from aita.tool.python_tool import PythonTool


class SparkTool(PythonTool):
    """Tool for running PySpark script."""

    name: str = "pyspark_tool"
    description: str = """
    Run a PySpark analysis script.
    The last line of the script should return a PySpark dataframe.
    If the script is not correct, an error message will be returned.
    """


class ConvertPandasToSparkTool(PythonTool):
    """Tool for converting a Pandas dataframe to a PySpark dataframe."""

    name: str = "convert_pandas_to_pyspark"
    description: str = """
    Convert a Pandas dataframe to a PySpark dataframe.
    If the Pandas dataframe is not correct, an error message will be returned.
    """

    def _run(self, dataframe: pd.DataFrame, **kwargs):
        """Convert the Pandas dataframe to a PySpark dataframe."""
        return self.script_context["spark_session"].createDataFrame(dataframe)
